from typing import List, Union, BinaryIO

from . import Object
from balethon import objects
from ..enums import MessageMediaType
from ..sync_support import add_sync_support_to_object


@add_sync_support_to_object
class Message(Object):
    attribute_names = [
        ("id", "message_id"),
        ("author", "from")
    ]

    @classmethod
    def wrap(cls, raw_object):
        try:
            if not raw_object.get("voice") and raw_object["document"]["mime_type"].startswith("audio"):
                raw_object["audio"] = raw_object["document"]
        except (TypeError, KeyError):
            pass
        try:
            if not raw_object.get("video") and raw_object["document"]["mime_type"].startswith("video"):
                raw_object["animation"] = raw_object["document"]
        except (TypeError, KeyError):
            pass
        return super().wrap(raw_object)

    def __init__(
            self,
            id: int = None,
            author: "objects.User" = None,
            date: "objects.Date" = None,
            chat: "objects.Chat" = None,
            forward_from: "objects.User" = None,
            forward_from_chat: "objects.Chat" = None,
            forward_from_message_id: int = None,
            forward_date: "objects.Date" = None,
            reply_to_message: "objects.Message" = None,
            edit_date: "objects.Date" = None,
            text: str = None,
            animation: "objects.Animation" = None,
            entities: List["objects.Entity"] = None,
            caption_entities: List["objects.Entity"] = None,
            voice: "objects.Voice" = None,
            audio: "objects.Audio" = None,
            document: "objects.Document" = None,
            photo: List["objects.Photo"] = None,
            video: "objects.Video" = None,
            sticker: "objects.Sticker" = None,
            caption: str = None,
            contact: "objects.Contact" = None,
            location: "objects.Location" = None,
            new_chat_members: List["objects.User"] = None,
            left_chat_member: "objects.User" = None,
            new_chat_title: str = None,
            new_chat_photo: List["objects.Photo"] = None,
            delete_chat_photo: bool = None,
            group_chat_created: bool = None,
            supergroup_chat_created: bool = None,
            channel_chat_created: bool = None,
            pinned_message: None = None,
            invoice: "objects.Invoice" = None,
            successful_payment: "objects.SuccessfulPayment" = None,
            media_group_id: int = None,
            web_app_data: "objects.WebAppData" = None,
            reply_markup: "objects.ReplyMarkup" = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.id: int = id
        self.author: "objects.User" = author
        self.date: "objects.Date" = date
        self.chat: "objects.Chat" = chat
        self.forward_from: "objects.User" = forward_from
        self.forward_from_chat: "objects.Chat" = forward_from_chat
        self.forward_from_message_id: int = forward_from_message_id
        self.forward_date: "objects.Date" = forward_date
        self.reply_to_message: "objects.Message" = reply_to_message
        self.edit_date: "objects.Date" = edit_date
        self.text: str = text
        self.animation: "objects.Animation" = animation
        self.entities: List["objects.Entity"] = entities
        self.caption_entities: List["objects.Entity"] = caption_entities
        self.voice: "objects.Voice" = voice
        self.audio: "objects.Audio" = audio
        self.document: "objects.Document" = document
        self.photo: List["objects.Photo"] = photo
        self.video: "objects.Video" = video
        self.sticker: "objects.Sticker" = sticker
        self.caption: str = caption
        self.contact: "objects.Contact" = contact
        self.location: "objects.Location" = location
        self.new_chat_members: List["objects.User"] = new_chat_members
        self.left_chat_member: "objects.User" = left_chat_member
        self.new_chat_title: str = new_chat_title
        self.new_chat_photo: List["objects.Photo"] = new_chat_photo
        self.delete_chat_photo: bool = delete_chat_photo
        self.group_chat_created: bool = group_chat_created
        self.supergroup_chat_created: bool = supergroup_chat_created
        self.channel_chat_created: bool = channel_chat_created
        self.pinned_message: None = pinned_message
        self.invoice: "objects.Invoice" = invoice
        self.successful_payment: "objects.SuccessfulPayment" = successful_payment
        self.media_group_id: int = media_group_id
        self.web_app_data: "objects.WebAppData" = web_app_data
        self.reply_markup: "objects.ReplyMarkup" = reply_markup

    @property
    def media_type(self):
        for media_type in MessageMediaType:
            if getattr(self, media_type.value):
                return media_type

    @property
    def content(self) -> str:
        return self.text or self.caption or ""

    @content.setter
    def content(self, value):
        if self.caption or self.document or self.contact or self.location or self.photo:
            self.caption = value
        else:
            self.text = value

    async def reply_animation(
            self,
            animation: Union[str, bytes, BinaryIO, "objects.InputMedia"],
            duration: int = None,
            width: int = None,
            height: int = None,
            caption: str = None,
            reply_markup: "objects.ReplyMarkup" = None
    ):
        return await self.client.send_animation(
            self.chat.id,
            animation,
            duration,
            width,
            height,
            caption,
            reply_markup,
            self.id
        )

    async def reply_audio(
            self,
            audio: Union[str, bytes, BinaryIO, "objects.InputMedia"],
            caption: str = None,
            duration: int = None,
            title: str = None,
            reply_markup: "objects.ReplyMarkup" = None
    ):
        return await self.client.send_audio(self.chat.id, audio, caption, duration, title, reply_markup, self.id)

    async def reply_contact(
            self,
            phone_number: str,
            first_name: str,
            last_name: str = None,
            reply_markup: "objects.ReplyMarkup" = None
    ):
        return await self.client.send_contact(self.chat.id, phone_number, first_name, last_name, reply_markup, self.id)

    async def reply_document(
            self,
            document: Union[str, bytes, BinaryIO, "objects.InputMedia"],
            caption: str = None,
            reply_markup: "objects.ReplyMarkup" = None
    ):
        return await self.client.send_document(self.chat.id, document, caption, reply_markup, self.id)

    async def reply_location(
            self,
            latitude: int,
            longitude: int,
            reply_markup: "objects.ReplyMarkup" = None
    ):
        return await self.client.send_location(self.chat.id, longitude, latitude, reply_markup, self.id)

    async def reply_media_group(
            self,
            media: List["objects.InputMedia"]
    ):
        return await self.client.send_media_group(self.chat.id, media)

    async def reply_photo(
            self,
            photo: Union[str, bytes, BinaryIO, "objects.InputMedia"],
            caption: str = None,
            reply_markup: "objects.ReplyMarkup" = None
    ):
        return await self.client.send_photo(self.chat.id, photo, caption, reply_markup, self.id)

    async def reply_sticker(
            self,
            sticker: Union[str, bytes, BinaryIO, "objects.InputMedia"],
            reply_markup: "objects.ReplyMarkup" = None
    ):
        return await self.client.send_sticker(self.chat.id, sticker, reply_markup, self.id)

    async def reply_video(
            self,
            video: Union[str, bytes, BinaryIO, "objects.InputMedia"],
            duration: int = None,
            width: int = None,
            height: int = None,
            caption: str = None,
            reply_markup: "objects.ReplyMarkup" = None
    ):
        return await self.client.send_video(self.chat.id, video, duration, width, height, caption, reply_markup)

    async def reply_voice(
            self,
            voice: Union[str, bytes, BinaryIO, "objects.InputMedia"],
            caption: str = None,
            duration: int = None,
            reply_markup: "objects.ReplyMarkup" = None
    ):
        return await self.client.send_voice(self.chat.id, voice, caption, duration, reply_markup, self.id)

    async def reply(
            self,
            text: str,
            reply_markup: "objects.ReplyMarkup" = None
    ):
        return await self.client.send_message(self.chat.id, text, reply_markup, self.id)

    async def edit_text(
            self,
            text: str,
            reply_markup: "objects.ReplyMarkup" = None
    ):
        return await self.client.edit_message_text(self.chat.id, self.id, text, reply_markup)

    async def edit_caption(
            self,
            caption: str,
            reply_markup: "objects.ReplyMarkup" = None
    ):
        return await self.client.edit_message_caption(self.chat.id, self.id, caption, reply_markup)

    async def edit(
            self,
            text: str,
            reply_markup: "objects.ReplyMarkup" = None
    ):
        if self.text:
            return await self.edit_text(text, reply_markup)
        return await self.edit_caption(text, reply_markup)

    async def edit_reply_markup(
            self,
            reply_markup: "objects.ReplyMarkup"
    ):
        return await self.client.edit_message_reply_markup(self.chat.id, self.id, reply_markup)

    async def delete(
            self
    ):
        return await self.client.delete_message(self.chat.id, self.id)

    async def forward(
            self,
            chat_id: Union[int, str]
    ):
        return await self.client.forward_message(chat_id, self.chat.id, self.id)

    async def copy(
            self,
            chat_id: Union[int, str]
    ):
        return await self.client.copy_message(chat_id, self.chat.id, self.id)

    async def send(
            self,
            chat_id: Union[int, str],
            reply_markup: "objects.ReplyMarkup" = None,
            reply_to_message_id: int = None
    ):
        if self.text:
            return await self.client.send_message(
                chat_id,
                self.text,
                reply_markup,
                reply_to_message_id
            )
        elif self.photo:
            return await self.client.send_photo(
                chat_id,
                self.photo[0].id,
                self.caption,
                reply_markup,
                reply_to_message_id
            )
        elif self.voice:
            return await self.client.send_voice(
                chat_id,
                self.voice.id,
                caption=self.caption,
                reply_markup=reply_markup,
                reply_to_message_id=reply_to_message_id
            )
        elif self.audio:
            return await self.client.send_audio(
                chat_id,
                self.audio.id,
                caption=self.caption,
                reply_markup=reply_markup,
                reply_to_message_id=reply_to_message_id
            )
        elif self.video:
            return await self.client.send_video(
                chat_id,
                self.video.id,
                caption=self.caption,
                reply_markup=reply_markup,
                reply_to_message_id=reply_to_message_id
            )
        elif self.animation:
            return await self.client.send_animation(
                chat_id,
                self.animation.id,
                caption=self.caption,
                reply_markup=reply_markup,
                reply_to_message_id=reply_to_message_id
            )
        elif self.contact:
            return await self.client.send_contact(
                chat_id,
                self.contact.phone_number,
                self.contact.first_name,
                self.contact.last_name,
                reply_markup=reply_markup,
                reply_to_message_id=reply_to_message_id
            )
        elif self.location:
            return await self.client.send_location(
                chat_id,
                self.location.longitude,
                self.location.latitude,
                reply_markup=reply_markup,
                reply_to_message_id=reply_to_message_id
            )
        elif self.sticker:
            return await self.client.send_sticker(
                chat_id,
                self.sticker.id,
                reply_markup=reply_markup,
                reply_to_message_id=reply_to_message_id
            )
        elif self.document:
            return await self.client.send_document(
                chat_id,
                self.document.id,
                self.caption,
                reply_markup,
                reply_to_message_id
            )
        raise TypeError("Message is not copyable")



from . import Object
from balethon import objects


class EditedMessage(Object):
    attribute_names = [
        ("id", "message_id"),
        ("author", "from")
    ]

    def __init__(
            self,
            id: int = None,
            author: "objects.User" = None,
            date: "objects.Date" = None,
            chat: "objects.Chat" = None,
            edit_date: "objects.Date" = None,
            text: str = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.id: int = id
        self.author: "objects.User" = author
        self.date: "objects.Date" = date
        self.chat: "objects.Chat" = chat
        self.edit_date: "objects.Date" = edit_date
        self.text: str = text



import balethon
from ...objects import Transaction


class InquireTransaction:

    async def inquire_transaction(
            self: "balethon.Client",
            transaction_id: str
    ) -> Transaction:
        return await self.auto_execute("post", "inquireTransaction", locals())



from typing import Union, BinaryIO

from . import InputMedia


class InputMediaVideo(InputMedia):

    def __init__(
            self,
            media: Union[str, bytes, BinaryIO] = None,
            caption: str = None,
            **kwargs
    ):
        super().__init__("video", media, caption, **kwargs)



from re import compile

from .condition import Condition
from ..objects import Message, CallbackQuery


class Regex(Condition):
    def __init__(self, pattern, flags=0):
        super().__init__(can_process=(Message, CallbackQuery))
        self.pattern = compile(pattern, flags)

    async def __call__(self, client, event) -> bool:
        if isinstance(event, Message):
            event = event.content
        elif isinstance(event, CallbackQuery):
            event = event.data
        return bool(list(self.pattern.finditer(event)) or None)



from typing import Union, List

from . import ReplyMarkup, ReplyKeyboardButton
from .list import Object, List as BalethonList
from balethon import objects


class ReplyKeyboard(ReplyMarkup):
    attribute_names = [
        ("resize", "resize_keyboard"),
        ("one_time", "one_time_keyboard"),
        ("remove", "remove_keyboard")
    ]

    def __init__(
            self,
            *rows: List[Union["objects.ReplyKeyboardButton", str]],
            resize: bool = None,
            one_time: bool = None,
            selective: bool = None,
            remove: bool = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        if "keyboard" not in kwargs:
            self.keyboard: List[List["objects.ReplyKeyboardButton"]] = BalethonList()
        for row in rows:
            self.add_row(*row)
        self.resize: bool = resize
        self.one_time: bool = one_time
        self.selective: bool = selective
        self.remove: bool = remove

    @classmethod
    def expected_types(cls):
        expected_types = super().expected_types()
        expected_types["keyboard"] = List[List[objects.ReplyKeyboardButton]]
        return expected_types

    @classmethod
    def wrap(cls, raw_object):
        return Object.wrap.__func__(cls, raw_object)

    def add_button(
            self,
            button: Union["objects.ReplyKeyboardButton", str],
            row_index: int = -1,
            button_index: int = -1
    ):
        if isinstance(button, str):
            button = ReplyKeyboardButton(button)
        if button_index == -1:
            self.keyboard[row_index].append(button)
        elif button_index < 0:
            self.keyboard[row_index].insert(button_index + 1, button)
        else:
            self.keyboard[row_index].insert(button_index, button)

    def add_row(self, *row: Union["objects.ReplyKeyboardButton", str], row_index: int = -1):
        if row_index == -1:
            self.keyboard.append(BalethonList())
        elif row_index < 0:
            self.keyboard.insert(row_index + 1, BalethonList())
        else:
            self.keyboard.insert(row_index, BalethonList())
        for button in row:
            self.add_button(button, row_index)

    def on_click(self, row_index: int, button_index: int):
        from ..conditions import click
        return click(self, row_index, button_index)



from . import Object


class Audio(Object):
    attribute_names = [
        ("id", "file_id"),
        ("size", "file_size")
    ]

    def __init__(
            self,
            id: str = None,
            duration: int = None,
            performer: str = None,
            title: str = None,
            mime_type: str = None,
            size: int = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.id: str = id
        self.duration: int = duration
        self.performer: str = performer
        self.title: str = title
        self.mime_type: str = mime_type
        self.size: int = size



from . import Object
from balethon import objects


class Sticker(Object):
    attribute_names = [
        ("id", "file_id"),
        ("unique_id", "file_unique_id"),
        ("size", "file_size")
    ]

    def __init__(
            self,
            id: str = None,
            unique_id: str = None,
            width: int = None,
            height: int = None,
            size: int = None,
            thumb: "objects.Photo" = None,
            emoji: str = None,
            set_name: str = None,
            type: str = None,
            is_animated: bool = None,
            is_video: bool = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.id: str = id
        self.unique_id: str = unique_id
        self.width: int = width
        self.height: int = height
        self.size: int = size
        self.thumb: "objects.Photo" = thumb
        self.emoji: str = emoji
        self.set_name: str = set_name
        self.type: str = type
        self.is_animated: bool = is_animated
        self.is_video: bool = is_video



from .event_handler import EventHandler
from .connect_handler import ConnectHandler
from .disconnect_handler import DisconnectHandler
from .error_handler import ErrorHandler
from .update_handler import UpdateHandler
from .message_handler import MessageHandler
from .callback_query_handler import CallbackQueryHandler
from .command_handler import CommandHandler
from .shipping_query_handler import ShippingQueryHandler
from .pre_checkout_query_handler import PreCheckoutQueryHandler
from .edited_message_handler import EditedMessageHandler
from .initialize_handler import InitializeHandler
from .shutdown_handler import ShutdownHandler



from . import Object


class Location(Object):

    def __init__(
            self,
            longitude: int = None,
            latitude: int = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.longitude: int = longitude
        self.latitude: int = latitude



from . import Object


class LabeledPrice(Object):

    def __init__(
            self,
            label: str = None,
            amount: int = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.label: str = label
        self.amount: int = amount



from . import Object


class WebAppInfo(Object):
    def __init__(
            self,
            url: str = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.url: str = url



from .object import Object


class ShippingAddress(Object):

    def __init__(
            self,
            country_code: str = None,
            stat: str = None,
            city: str = None,
            street_line1: str = None,
            street_line2: str = None,
            post_code: str = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.country_code: str = country_code
        self.stat: str = stat
        self.city: str = city
        self.street_line1: str = street_line1
        self.street_line2: str = street_line2
        self.post_code: str = post_code



from sys import stdout

from .chain import Chain


class MonitoringChain(Chain):
    def __init__(self, name="monitoring", condition=None, stream=None):
        super().__init__(name, condition)
        self.stream = stream or stdout

    @Chain.message_handler()
    def monitor_message(self, message):
        if not message.author:
            return
        self.stream.write(f"{message.author.full_name}: {message.content}\n")

    @Chain.callback_query_handler()
    def monitor_callback_query(self, callback_query):
        self.stream.write(f"{callback_query.author.full_name}: [{callback_query.data}]\n")



from . import Object


class Invoice(Object):

    def __init__(
            self,
            title: str = None,
            description: str = None,
            start_parameter: str = None,
            currency: str = None,
            total_amount: int = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.title: str = title
        self.description: str = description
        self.start_parameter: str = start_parameter
        self.currency: str = currency
        self.total_amount: int = total_amount



from inspect import getfullargspec, signature


def remove_unwanted_positional_parameters(function, *args):
    _, varargs, *__ = getfullargspec(function)

    if varargs is not None:
        return args

    positional_arguments = []

    for i, parameter in enumerate(signature(function).parameters.values()):
        if parameter.kind in (parameter.POSITIONAL_OR_KEYWORD, parameter.POSITIONAL_ONLY):
            try:
                positional_arguments.append(args[i])
            except IndexError:
                continue

    return positional_arguments


def remove_unwanted_keyword_parameters(function, **kwargs):
    _, __, varkw, *___ = getfullargspec(function)

    if varkw is not None:
        return kwargs

    keyword_arguments = {}

    for name, parameter in signature(function).parameters.items():
        if parameter.kind in (parameter.POSITIONAL_OR_KEYWORD, parameter.KEYWORD_ONLY):
            try:
                keyword_arguments[name] = kwargs[name]
            except KeyError:
                continue

    return keyword_arguments


def remove_unwanted_parameters(function, *args, **kwargs):
    args = remove_unwanted_positional_parameters(function, *args)
    kwargs = remove_unwanted_keyword_parameters(function, **kwargs)

    return args, kwargs



class RPCError(Exception):
    name = "RPC Error"
    code = None

    @classmethod
    def create(cls, code: int = None, description: str = None, reason: str = None, parameters: dict = None):
        for rpc_error in cls.__subclasses__():
            if code == rpc_error.code:
                return rpc_error(code, description, reason, parameters)
        return cls(code, description, reason, parameters)

    def __init__(self, code: int = None, description: str = None, reason: str = None, parameters: dict = None):
        self.code = code or self.code
        self.description = description
        self.reason = reason
        self.parameters = parameters
        super().__init__(str(self))

    def __str__(self):
        code = "" if self.code is None else f" {self.code}"
        description = "" if self.description is None else f": {self.description}"
        reason = "" if self.reason is None else f" (caused by {self.reason})"
        return f"{self.name}{code}{description}{reason}"



from typing import Union
from json import dumps
from typing import List

import balethon
from ...objects import Message, ReplyMarkup
from balethon import objects


class SendMediaGroup:

    async def send_media_group(
            self: "balethon.Client",
            chat_id: Union[int, str],
            media: List["objects.InputMedia"],
            reply_markup: ReplyMarkup = None,
            reply_to_message_id: int = None
    ) -> List[Message]:
        chat_id = await self.resolve_peer_id(chat_id)
        data = locals()
        del data["self"]
        if any(not m.is_json_serializable for m in data["media"]):
            for i, m in enumerate(data["media"]):
                data["media"][i] = m.unwrap()
                if not m.is_json_serializable:
                    data[f"media{i}"] = m.media
                    data["media"][i]["media"] = f"media{i}"
            data["media"] = dumps(data["media"])
        else:
            for i, m in enumerate(data["media"]):
                data["media"][i] = m.unwrap()
        result = await self.execute("post", "sendMediaGroup", **data)
        result = [Message.wrap(message) for message in result]
        for message in result:
            message.bind(self)
        return result



from enum import Enum


class NameEnum(Enum):
    @staticmethod
    def _generate_next_value_(name, start, count, last_values):
        return name.lower()

    def __repr__(self):
        return super().__str__()

    def __eq__(self, other):
        return self.value == other



from typing import Union

import balethon
from ...objects import Message


class CopyMessage:

    async def copy_message(
            self: "balethon.Client",
            chat_id: Union[int, str],
            from_chat_id: Union[int, str],
            message_id: int
    ) -> Message:
        chat_id = await self.resolve_peer_id(chat_id)
        from_chat_id = await self.resolve_peer_id(from_chat_id)
        return await self.auto_execute("post", "copyMessage", locals())



from typing import List

from . import Object
from balethon import objects


class StickerSet(Object):

    def __init__(
            self,
            name: str = None,
            title: str = None,
            sticker_type: str = None,
            is_animated: bool = None,
            is_video: bool = None,
            contains_masks: bool = None,
            stickers: List["objects.Sticker"] = None,
            thumb: "objects.Photo" = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.name: str = name
        self.title: str = title
        self.sticker_type: str = sticker_type
        self.is_animated: bool = is_animated
        self.is_video: bool = is_video
        self.contains_masks: bool = contains_masks
        self.stickers: List["objects.Sticker"] = stickers
        self.thumb: "objects.Photo" = thumb



from . import Object
from balethon import objects


class Document(Object):
    attribute_names = [
        ("id", "file_id"),
        ("size", "file_size"),
        ("name", "file_name")
    ]

    def __init__(
            self,
            id: str = None,
            thumb: "objects.Photo" = None,
            name: str = None,
            mime_type: str = None,
            size: int = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.id: str = id
        self.thumb: None = thumb
        self.name: str = name
        self.mime_type: str = mime_type
        self.size: int = size



from typing import List

import balethon
from ...objects import Update


class GetUpdates:

    async def get_updates(
            self: "balethon.Client",
            offset: int = None,
            limit: int = None
    ) -> List[Update]:
        return await self.auto_execute("post", "getUpdates", locals())



from balethon.objects import Object


class CopyTextButton(Object):
    def __init__(
            self,
            text: str = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.text: str = text



from typing import Union, List

import balethon
from ...objects import Message, LabeledPrice
from balethon import objects


class SendInvoice:

    async def send_invoice(
            self: "balethon.Client",
            chat_id: Union[int, str],
            title: str,
            description: str,
            payload: str,
            provider_token: str,
            prices: List[LabeledPrice],
            provider_data: str = None,
            photo_url: str = None,
            photo_size: int = None,
            photo_width: int = None,
            photo_height: int = None,
            need_name: bool = None,
            need_phone_number: bool = None,
            need_email: bool = None,
            need_shipping_address: bool = None,
            is_flexible: bool = None,
            disable_notification: bool = None,
            reply_to_message_id: int = None,
            reply_markup: "objects.ReplyMarkup" = None
    ) -> Message:
        chat_id = await self.resolve_peer_id(chat_id)
        return await self.auto_execute("post", "sendInvoice", locals())



from . import Object


class Photo(Object):
    attribute_names = [
        ("id", "file_id"),
        ("size", "file_size")
    ]

    def __init__(
            self,
            id: str = None,
            width: int = None,
            height: int = None,
            size: int = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.id: str = id
        self.width: int = width
        self.height: int = height
        self.size: int = size



from . import Object


class Transaction(Object):
    attribute_names = [
        ("user_id", "userID"),
        ("created_at", "createdAt"),
    ]

    def __init__(
            self,
            id: str = None,
            status: str = None,
            user_id: int = None,
            amount: int = None,
            created_at: int = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.id: str = id
        self.status: str = status
        self.user_id: int = user_id
        self.amount: int = amount
        self.created_at: int = created_at



from typing import Union, Tuple

from .condition import Condition
from ..objects import Message, CallbackQuery
from ..enums import ChatMemberStatus
from ..errors import RPCError, ForbiddenError


class IsJoined(Condition):
    ACCEPTED_STATUSES = ChatMemberStatus.MEMBER, ChatMemberStatus.ADMINISTRATOR, ChatMemberStatus.CREATOR

    def __init__(self, *chat_ids: Union[int, str], accepted_statuses: Tuple[str, ...] = None):
        super().__init__(can_process=(Message, CallbackQuery))
        self.chat_ids = list(chat_ids)
        self.chats = []
        self.accepted_statuses = accepted_statuses or self.ACCEPTED_STATUSES

    async def update_chats(self, client):
        self.chats = []
        for chat_id in self.chat_ids:
            chat = await client.get_chat(chat_id)
            self.chats.append(chat)
        self.chat_ids = [chat.id for chat in self.chats]

    async def add_chat(self, client, chat_id):
        chat = await client.get_chat(chat_id)
        self.chats.append(chat)
        self.chat_ids.append(chat.id)
        return chat

    async def remove_chat(self, client, chat_id):
        chat = await client.get_chat(chat_id)
        for c in self.chats:
            if c.id == chat.id:
                index = self.chats.index(c)
                self.chats.pop(index)
        self.chat_ids.remove(chat.id)
        return chat

    async def __call__(self, client, event) -> bool:
        if not self.chats:
            await self.update_chats(client)
        if not event.author:
            return False
        event.not_joined_chats = []
        for chat in self.chats:
            try:
                chat_member = await client.get_chat_member(chat.id, event.author.id)
            except ForbiddenError as error:
                raise error
            except RPCError:
                event.not_joined_chats.append(chat)
            else:
                if chat_member.status not in self.accepted_statuses:
                    event.not_joined_chats.append(chat)
        return not bool(event.not_joined_chats)



from .delete_sticker_from_set import DeleteStickerFromSet
from .get_sticker_set import GetStickerSet
from .create_new_sticker_set import CreateNewStickerSet
from .upload_sticker_file import UploadStickerFile
from .add_sticker_to_set import AddStickerToSet


class Stickers(GetStickerSet, CreateNewStickerSet, DeleteStickerFromSet, UploadStickerFile, AddStickerToSet):
    pass



from traceback import print_exception

from .chain import Chain


class PrintingChain(Chain):

    def __init__(self, name="printing", condition=None):
        super().__init__(name, condition)

    @Chain.error_handler()
    def print_error(self, error):
        print_exception(None, error, error.__traceback__)

    @Chain.initialize_handler()
    def print_ready(self, client):
        print(f"---{client} is ready---")

    @Chain.shutdown_handler()
    def print_stopped(self, client):
        print(f"---{client} has stopped---")



from asyncio import get_event_loop, new_event_loop, Queue
from concurrent.futures import ThreadPoolExecutor
from os import cpu_count

from .chain import Chain
from ..errors import ContinueDispatching, BreakDispatching
from ..event_handlers import EventHandler
from ..objects import Update


def is_subclass(cls, super_cls):
    if not isinstance(cls, type):
        return False
    return issubclass(cls, super_cls)


class Dispatcher:
    reasonable_sync_workers = min(32, (cpu_count() or 1) + 4)
    reasonable_async_workers = reasonable_sync_workers * 5

    def __init__(
            self,
            *clients,
            async_workers: int = None,
            sync_workers: int = None
    ):
        self.clients = clients
        try:
            self.event_loop = get_event_loop()
        except RuntimeError:
            self.event_loop = new_event_loop()
        self.events_queue = Queue()
        self.is_started = False
        self.async_workers_count = self.reasonable_async_workers or async_workers
        self.async_workers = []
        self.sync_workers_count = self.reasonable_sync_workers or sync_workers
        self.sync_workers = None

    async def start(self):
        if self.is_started:
            raise RuntimeError("Dispatcher is already started")
        self.is_started = True
        self.async_workers = [self.event_loop.create_task(self.worker()) for _ in range(self.async_workers_count)]
        self.sync_workers = ThreadPoolExecutor(self.sync_workers_count, thread_name_prefix="Event Handler")

    async def stop(self):
        if not self.is_started:
            raise RuntimeError("Dispatcher is already stopped")
        self.is_started = False
        for _ in self.async_workers:
            self.events_queue.put_nowait(None)
        for async_worker in self.async_workers:
            await async_worker
        self.async_workers.clear()
        self.sync_workers.shutdown()

    async def propagate_chain(self, chain, client, event):
        for child in chain.children:
            if isinstance(child, Chain):
                if not await child.check(client, event):
                    continue
                await self.propagate_chain(child, client, event)
                break
            elif isinstance(child, EventHandler):
                if not isinstance(event, child.can_handle) and not is_subclass(event, child.can_handle):
                    continue
                try:
                    if not await child.check(client, event):
                        continue
                    await child.handle(client=client, event=event)
                    break
                except ContinueDispatching:
                    continue
                except BreakDispatching:
                    break
                except Exception as error:
                    await self.dispatch_event(client, error)
                    break
        for child in chain.chains:
            if not await child.check(client, event):
                continue
            await self.propagate_chain(child, client, event)

    async def worker(self):
        while True:
            item = await self.events_queue.get()
            if item is None:
                break
            client, event = item
            await self.propagate_chain(client, client, event)

    async def dispatch_event(self, client, event):
        if self.is_started:
            self.events_queue.put_nowait((client, event))

    async def dispatch_raw_update(self, client, update):
        update = Update.wrap(update).get_effective_update()
        update.bind(client)
        await self.dispatch_event(client, update)



from .condition import create
from ..objects import Message


@create(can_process=Message)
def number(event) -> bool:
    try:
        int(event.content)
    except ValueError:
        return False
    else:
        return True



from .condition import Condition
from ..objects import Message, CallbackQuery
from ..states import State


class AtState(Condition):
    def __init__(self, state, state_machine=None):
        super().__init__(can_process=(Message, CallbackQuery))
        self.state = str(state) if isinstance(state, State) else state
        self.state_machine = state_machine

    async def __call__(self, client, event) -> bool:
        if not event.author:
            return False
        if self.state_machine is None:
            return event.author.get_state() == self.state
        else:
            return self.state_machine[event.author.id] == self.state



from .bad_request_error import BadRequestError
from .flood_error import FloodError
from .forbidden_error import ForbiddenError
from .internal_error import InternalError
from .not_found_error import NotFoundError
from .rpc_error import RPCError
from .too_many_requests import TooManyRequestsError
from .unauthorized_error import UnauthorizedError



from . import Object


class Contact(Object):

    def __init__(
            self,
            phone_number: str = None,
            first_name: str = None,
            last_name: str = None,
            user_id: int = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.phone_number: str = phone_number
        self.first_name: str = first_name
        self.last_name: str = last_name
        self.user_id: int = user_id



from .answer_pre_checkout_query import AnswerPreCheckoutQuery
from .inquire_transacion import InquireTransaction
from .send_invoice import SendInvoice


class Payments(AnswerPreCheckoutQuery, InquireTransaction, SendInvoice):
    pass



from . import State


class StateGroup:

    @classmethod
    def get_state_dict(cls):
        return {name: state for name, state in cls.__dict__.items() if isinstance(state, State)}

    def __init_subclass__(cls, **kwargs):
        for name, state in cls.get_state_dict().items():
            state.name = name
            state.group = cls



from typing import Union

import balethon
from ...objects import Chat
from ...errors import RPCError
from ...enums import ChatType


class GetChat:

    async def get_chat(
            self: "balethon.Client",
            chat_id: Union[int, str]
    ) -> Chat:
        # 1234567890 | "1234567890"
        if isinstance(chat_id, int) or (isinstance(chat_id, str) and chat_id.isnumeric()):
            return await self.auto_execute("post", "getChat", locals())

        # "@username"
        if chat_id.startswith("@"):
            chat_id = chat_id.lstrip("@")

        # "+9*********" | "+09*********" | "+989*********"
        elif chat_id.startswith("+"):
            chat_id = "@" + chat_id.lstrip("+")

        # "https://ble.ir/join/ABCDEEFGHI" | "https://ble.ir/username"
        elif chat_id.startswith(f"{self.connection.SHORT_URL}/"):
            chat_id = chat_id.replace(f"{self.connection.SHORT_URL}/", "")

        # "ble.ir/join/ABCDEEFGHI" | "ble.ir/username"
        elif chat_id.startswith("ble.ir/"):
            chat_id = chat_id.replace("ble.ir/", "")

        ver, info = await self.connection.get_peer_info(chat_id)
        result = Chat()
        
        if ver == "v1":
            if not info.get("ok", False):
                raise RPCError.create(code=404, description="no such group or user", reason="getChat")

            info = info.get("result", {})
            result.type = ChatType.CHANNEL if info.get("type", "") == "channel" else ChatType.GROUP
            result.username = info.get("username", "")
            result.title = info.get("title", "")
            result.description = info.get("description") or info.get("bio") or ""
            result.invite_link = info.get("invite_link", "")
            result.id = info.get("id", "")

            result.bind(self)
            return result
        else:
            if info["query"].get("token"):
                token = info["query"]["token"]
                result.invite_link = f"{self.connection.SHORT_URL}/join/{token}"

            info = info["props"]["pageProps"]

            if info["peer"]["type"] == 0:
                raise RPCError.create(code=404, description="no such group or user", reason="getChat")

            elif info["peer"]["type"] == 1:
                result.type = ChatType.PRIVATE
                result.username = info["user"]["nick"] if info["user"].get("nick") else None
                result.first_name = info["user"]["title"]
                result.description = info["user"]["description"]

            elif info["peer"]["type"] == 2:
                result.type = ChatType.CHANNEL if info["group"]["isChannel"] else ChatType.GROUP
                result.username = info["group"]["nick"] if info["group"].get("nick") else None
                result.title = info["group"]["title"]
                result.description = info["group"]["description"]

            result.id = info["peer"]["id"]
            result.bind(self)
            return result



import balethon


class DeleteStickerFromSet:

    async def delete_sticker_from_set(
            self: "balethon.Client",
            sticker: str
    ) -> bool:
        return await self.auto_execute("get", "deleteStickerToSet", locals())



from typing import Union

import balethon


class DeleteMessage:

    async def delete_message(
            self: "balethon.Client",
            chat_id: Union[int, str],
            message_id: int
    ) -> bool:
        chat_id = await self.resolve_peer_id(chat_id)
        return await self.auto_execute("get", "deleteMessage", locals())



from typing import Union, BinaryIO
from os.path import isfile

from . import Object


class InputMedia(Object):

    def __init__(
            self,
            type: str = None,
            media: Union[str, bytes, BinaryIO] = None,
            caption: str = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.type: str = type
        self.media: Union[str, bytes, BinaryIO] = resolve_media(media)
        self.caption: str = caption

    @property
    def is_json_serializable(self):
        return isinstance(self.media, str)


def resolve_media(media):
    if isinstance(media, InputMedia):
        return media.media

    if isinstance(media, Object):
        if getattr(media, "id", None):
            return media.id

    try:
        if isfile(media):
            return open(media, "rb")
    except TypeError:
        pass

    return media



from typing import Union

import balethon
from ...objects import Message
from balethon import objects


class EditMessageText:

    async def edit_message_text(
            self: "balethon.Client",
            chat_id: Union[int, str],
            message_id: int,
            text: str,
            reply_markup: "objects.ReplyMarkup" = None
    ) -> Message:
        chat_id = await self.resolve_peer_id(chat_id)
        return await self.auto_execute("post", "editMessageText", locals())



from typing import Union

from . import Object
from balethon import objects


class ReplyKeyboardButton(Object):

    def __init__(
            self,
            text: str = None,
            request_contact: bool = None,
            request_location: bool = None,
            web_app: Union["objects.WebAppInfo", str] = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.text: str = text
        self.request_contact: bool = request_contact
        self.request_location: bool = request_location
        self.web_app = objects.WebAppInfo(web_app) if isinstance(web_app, str) else web_app



from sqlite3 import connect
from typing import Union

from .state import State


class StateMachine:
    global_state_machine = None

    def create_table(self):
        sql = """CREATE TABLE IF NOT EXISTS user_states(user_id INTEGER PRIMARY KEY, user_state TEXT)"""
        cursor = self.connection.cursor()
        cursor.execute(sql)
        self.connection.commit()

    def __init__(self, database: str = ":memory:", state_group=None):
        self.connection = connect(database, check_same_thread=False)
        self.state_group = state_group
        self.create_table()

    def insert_user_state(self, user_id: Union[int, str], state: Union[State, str]):
        sql = """INSERT INTO user_states VALUES (?, ?)"""
        cursor = self.connection.cursor()
        cursor.execute(sql, (int(user_id), state))
        self.connection.commit()

    def update_user_state(self, user_id: Union[int, str], state: Union[State, str]):
        sql = """UPDATE user_states SET user_state = ? WHERE user_id = ?"""
        cursor = self.connection.cursor()
        cursor.execute(sql, (state, int(user_id)))
        self.connection.commit()

    def __setitem__(self, user_id: Union[int, str], state: Union[State, str]):
        if isinstance(state, State):
            state = str(state)
        if self[user_id] is None:
            self.insert_user_state(user_id, state)
        else:
            self.update_user_state(user_id, state)

    def select_user_state(self, user_id: Union[int, str]):
        sql = """SELECT user_state FROM user_states WHERE user_id = ?"""
        cursor = self.connection.cursor()
        cursor.execute(sql, (int(user_id),))
        return cursor.fetchone()

    def __getitem__(self, user_id: Union[int, str]):
        result = self.select_user_state(user_id)
        if result is None:
            return result
        elif self.state_group is not None and result[0] in self.state_group.get_state_dict():
            return getattr(self.state_group, result[0])
        else:
            return result[0]

    def delete_user_state(self, user_id: Union[int, str]):
        sql = """DELETE from user_states where user_id = ?"""
        cursor = self.connection.cursor()
        cursor.execute(sql, (int(user_id),))
        self.connection.commit()

    def __delitem__(self, user_id: Union[int, str]):
        if self[user_id] is None:
            raise KeyError("User does not exist")
        self.delete_user_state(user_id)

    def at(self, state: Union[State, str]):
        from ..conditions import at_state
        return at_state(state, self)

    def change_database(self, database: str = ":memory:"):
        now_connection = connect(database, check_same_thread=False)
        self.connection.backup(now_connection)
        self.connection = now_connection

    def get_all(self):
        sql = """SELECT * FROM user_states"""
        cursor = self.connection.cursor()
        cursor.execute(sql)
        return dict(cursor.fetchall())


StateMachine.global_state_machine = StateMachine()



from enum import auto

from .name_enum import NameEnum


class ChatAction(NameEnum):
    TYPING = auto()
    UPLOAD_PHOTO = auto()
    RECORD_VIDEO = auto()
    UPLOAD_VIDEO = auto()
    CHOOSE_STICKER = auto()
    UPLOAD_VOICE = auto()
    RECORD_VOICE = auto()



from typing import Union, List, Tuple

from . import Object, ReplyMarkup, InlineKeyboardButton
from .list import List as BalethonList
from balethon import objects


class InlineKeyboard(ReplyMarkup):
    def __init__(
            self,
            *rows: List[Union["objects.InlineKeyboardButton", Tuple[str, str]]],
            **kwargs
    ):
        super().__init__(**kwargs)
        if "inline_keyboard" not in kwargs:
            self.inline_keyboard: List[List["objects.InlineKeyboardButton"]] = BalethonList()
        for row in rows:
            self.add_row(*row)

    @classmethod
    def expected_types(cls):
        expected_types = super().expected_types()
        expected_types["inline_keyboard"] = List[List[objects.InlineKeyboardButton]]
        return expected_types

    @classmethod
    def wrap(cls, raw_object):
        return Object.wrap.__func__(cls, raw_object)

    def add_button(
            self,
            button: Union["objects.InlineKeyboardButton", Tuple[str, str]],
            row_index: int = -1,
            button_index: int = -1
    ):
        if isinstance(button, tuple):
            button = InlineKeyboardButton(*button)
        if button_index == -1:
            self.inline_keyboard[row_index].append(button)
        elif button_index < 0:
            self.inline_keyboard[row_index].insert(button_index + 1, button)
        else:
            self.inline_keyboard[row_index].insert(button_index, button)

    def add_row(self, *row: Union["objects.InlineKeyboardButton", Tuple[str, str]], row_index: int = -1):
        if row_index == -1:
            self.inline_keyboard.append(BalethonList())
        elif row_index < 0:
            self.inline_keyboard.insert(row_index + 1, BalethonList())
        else:
            self.inline_keyboard.insert(row_index, BalethonList())
        for button in row:
            self.add_button(button, row_index)

    def on_click(self, row_index: int, button_index: int):
        from ..conditions import click
        return click(self, row_index, button_index)



from typing import Union

import balethon


class UnpinChatMessage:

    async def unpin_chat_message(
            self: "balethon.Client",
            chat_id: Union[int, str],
            message_id: int = None
    ) -> bool:
        chat_id = await self.resolve_peer_id(chat_id)
        return await self.auto_execute("post", "unpinChatMessage", locals())



from json import dumps
from asyncio import get_event_loop, sleep
from inspect import iscoroutine, iscoroutinefunction, stack
from io import BufferedReader, BytesIO
from typing import get_type_hints

from httpx import ConnectError

from .messages import Messages
from .updates import Updates
from .users import Users
from .attachments import Attachments
from .chats import Chats
from .invite_links import InviteLinks
from .payments import Payments
from .stickers import Stickers
from ..objects import Object, wrap, unwrap, Chat, User
from ..errors import TooManyRequestsError
from ..network import Connection
from ..dispatcher import Dispatcher, Chain, PrintingChain
from ..event_handlers import ConnectHandler, DisconnectHandler, InitializeHandler, ShutdownHandler
from ..smart_call import remove_unwanted_keyword_parameters
from ..sync_support import add_sync_support_to_object


@add_sync_support_to_object
class Client(Chain, Messages, Updates, Users, Attachments, Chats, InviteLinks, Payments, Stickers):

    def __init__(
            self,
            token: str,
            async_workers: int = None,
            sync_workers: int = None,
            time_out: int = None,
            sleep_threshold: int = 60,
            proxy=None,
            base_url: str = None,
            short_url: str = None
    ):
        super().__init__("default", None, PrintingChain())
        self.dispatcher = Dispatcher(self, async_workers=async_workers, sync_workers=sync_workers)
        self.connection = Connection(token, time_out, proxy, base_url, short_url)
        self.sleep_threshold = sleep_threshold
        self.user = None
        self.is_disconnected = False

    def __repr__(self):
        client_name = type(self).__name__
        try:
            name = self.user.full_name
        except AttributeError:
            name = "Not initialized yet"
        return f"{client_name}({name})"

    async def connect(self):
        await self.connection.start()
        self.user = await self.get_me()

    async def disconnect(self):
        await self.connection.stop()

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, *args):
        try:
            await self.disconnect()
        except ConnectionError:
            return

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, *args):
        try:
            self.disconnect()
        except ConnectionError:
            return

    async def execute(self, method: str, service: str, json: bool = None, **data):
        data = {k: v for k, v in data.items() if v is not None}
        files = {}
        if json is None:
            for value in data.values():
                if isinstance(value, (bytes, BufferedReader, BytesIO)):
                    json = False
                    break
            else:
                json = True
        for key, value in data.items():
            data[key] = unwrap(value)
        if not json:
            for key, value in data.copy().items():
                if isinstance(value, (bytes, BufferedReader, BytesIO)):
                    files[key] = value
                    del data[key]
                elif isinstance(value, dict):
                    data[key] = dumps(value)
        while True:
            try:
                if json:
                    return await self.connection.request(method, service, json=data)
                return await self.connection.request(method, service, data=data, files=files)
            except TooManyRequestsError as error:
                if error.seconds <= self.sleep_threshold:
                    print(f"[Too many requests] retry after: {error.seconds} (caused by {service})")
                    await sleep(error.seconds)
                else:
                    raise error

    async def auto_execute(self, method: str, service: str, data: dict, json: bool = None):
        bound_method_name = stack()[1].function
        bound_method = getattr(self, bound_method_name)
        type_hints = get_type_hints(bound_method)
        del data["self"]
        del type_hints["self"]
        return_type_hint = type_hints.pop("return")
        result = await self.execute(method, service, json, **data)
        result = wrap(return_type_hint, result)
        if isinstance(result, Object):
            result.bind(self)
        return result

    async def initialize(self):
        await self.dispatcher.start()
        await self.dispatcher.dispatch_event(self, InitializeHandler)

    async def shutdown(self):
        await self.dispatcher.dispatch_event(self, ShutdownHandler)
        await self.dispatcher.stop()

    async def start_polling(self):
        await self.delete_webhook()
        await self.initialize()
        last_update_id = None
        while True:
            try:
                if last_update_id is None:
                    updates = await self.get_updates()
                else:
                    updates = await self.get_updates(last_update_id + 1)
            except ConnectError:
                if not self.is_disconnected:
                    self.is_disconnected = True
                    await self.dispatcher.dispatch_event(self, DisconnectHandler)
            except Exception as error:
                await self.dispatcher.dispatch_event(self, error)
            else:
                if self.is_disconnected:
                    self.is_disconnected = False
                    await self.dispatcher.dispatch_event(self, ConnectHandler)
                for update in updates:
                    if last_update_id is not None and last_update_id >= update.id:
                        continue
                    last_update_id = update.id
                    await self.dispatcher.dispatch_event(self, update.get_effective_update())

    def run(self, function=None):
        try:
            self.connect()
            if function is None:
                self.start_polling()
            elif iscoroutine(function):
                loop = get_event_loop()
                loop.run_until_complete(function)
            elif iscoroutinefunction(function):
                kwargs = remove_unwanted_keyword_parameters(function, client=self)
                loop = get_event_loop()
                loop.run_until_complete(function(**kwargs))
            else:
                kwargs = remove_unwanted_keyword_parameters(function, client=self)
                function(**kwargs)
        except KeyboardInterrupt:
            return
        finally:
            if self.dispatcher.is_started:
                self.shutdown()
            self.disconnect()

    async def download(self, file_id: str):
        return await self.connection.download_file(file_id)

    async def resolve_peer_id(self, chat_id):
        if isinstance(chat_id, str) and not chat_id.isnumeric():
            peer = await self.get_chat(chat_id)
            return peer.id
        if isinstance(chat_id, (Chat, User)):
            return chat_id.id
        return chat_id

    def create_referral_link(self, name, value) -> str:
        return f"{self.connection.short_url}/{self.user.username}?{name}={value}"



from typing import Union

from . import Object
from ..states import StateMachine
from ..sync_support import add_sync_support_to_object


@add_sync_support_to_object
class User(Object):

    def __init__(
            self,
            id: int = None,
            username: str = None,
            first_name: str = None,
            last_name: str = None,
            language_code: str = None,
            is_bot: bool = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.id: int = id
        self.username: str = username
        self.first_name: str = first_name
        self.last_name: str = last_name
        self.language_code: str = language_code
        self.is_bot: bool = is_bot

    def mention(self, text=None):
        text = text or self.full_name
        return f"[{text}](https://web.bale.ai/chat?uid={self.id})"

    @property
    def link(self):
        if self.username:
            return f"{self.client.connection.short_url}/{self.username}"

    @property
    def full_name(self):
        if self.first_name and self.last_name:
            return f"{self.first_name} {self.last_name}"
        if self.first_name:
            return self.first_name
        return ""

    def set_state(self, state):
        StateMachine.global_state_machine[self.id] = state

    def get_state(self):
        return StateMachine.global_state_machine[self.id]

    def del_state(self):
        del StateMachine.global_state_machine[self.id]

    async def invite(self, chat_id: Union[int, str]):
        return await self.client.invite_user(chat_id, self.id)



from .condition import Condition
from ..objects import Message


class Command(Condition):
    def __init__(self, name, prefix="/", min_arguments=None, max_arguments=None):
        super().__init__(can_process=Message)
        self.name = name
        self.prefix = prefix
        self.min_arguments = min_arguments
        self.max_arguments = max_arguments

    async def __call__(self, client, event) -> bool:
        if not event.text:
            return False
        name, *arguments = event.text.split()
        if name.lower() != f"{self.prefix}{self.name.lower()}":
            return False
        if self.min_arguments is not None and len(arguments) < self.min_arguments:
            return False
        if self.max_arguments is not None and len(arguments) > self.max_arguments:
            return False
        return True



class State:

    def __init__(self):
        self.name = None
        self.group = None

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"{self.group.__name__}.{self.name}"



from . import Object
from balethon import objects


class ShippingQuery(Object):
    attribute_names = [
        ("author", "from")
    ]

    def __init__(
            self,
            id: str = None,
            author: "objects.User" = None,
            invoice_payload: str = None,
            shipping_address: "objects.ShippingAddress" = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.id: str = id
        self.author: "objects.User" = author
        self.invoice_payload: str = invoice_payload
        self.shipping_address: "objects.ShippingAddress" = shipping_address



from ..event_handlers import (
    ConnectHandler,
    InitializeHandler,
    EventHandler,
    ErrorHandler,
    UpdateHandler,
    MessageHandler,
    EditedMessageHandler,
    CommandHandler,
    CallbackQueryHandler,
    PreCheckoutQueryHandler,
    ShippingQueryHandler,
    ShutdownHandler,
    DisconnectHandler
)


class Chain:

    def __init__(self, name, condition=None, *chains):
        self.name = name
        self.condition = condition
        self.chains = list(chains)
        self.children = []
        for attribute_name in dir(self):
            attribute = getattr(self, attribute_name)
            if isinstance(attribute, EventHandler):
                attribute.self = self
                self.add_event_handler(attribute)

    async def check(self, client, event):
        if self.condition is None:
            return True
        return await self.condition(client, event)

    @staticmethod
    def create_event_handler(event_handler, *args, **kwargs):
        def decorator(callback):
            return event_handler(callback, *args, **kwargs)
        return decorator

    def add_event_handler(self, event_handler, *args, **kwargs):
        if isinstance(event_handler, type):
            def decorator(callback):
                event_handler_instance = event_handler(callback, *args, **kwargs)
                self.add_event_handler(event_handler_instance)
                return event_handler_instance
            return decorator
        self.children.append(event_handler)

    @classmethod
    def connect_handler(cls):
        return cls.create_event_handler(ConnectHandler)

    def on_connect(self):
        return self.add_event_handler(ConnectHandler)

    @classmethod
    def initialize_handler(cls):
        return cls.create_event_handler(InitializeHandler)

    def on_initialize(self):
        return self.add_event_handler(InitializeHandler)

    @classmethod
    def event_handler(cls, condition=None):
        return cls.create_event_handler(EventHandler, condition)

    def on_event(self, condition=None):
        return self.add_event_handler(EventHandler, condition)

    @classmethod
    def error_handler(cls, condition=None):
        return cls.create_event_handler(ErrorHandler, condition)

    def on_error(self, condition=None):
        return self.add_event_handler(ErrorHandler, condition)

    @classmethod
    def update_handler(cls, condition=None):
        return cls.create_event_handler(UpdateHandler, condition)

    def on_update(self, condition=None):
        return self.add_event_handler(UpdateHandler, condition)

    @classmethod
    def message_handler(cls, condition=None):
        return cls.create_event_handler(MessageHandler, condition)

    def on_message(self, condition=None):
        return self.add_event_handler(MessageHandler, condition)

    @classmethod
    def edited_message_handler(cls, condition=None):
        return cls.create_event_handler(EditedMessageHandler, condition)

    def on_edited_message(self, condition=None):
        return self.add_event_handler(EditedMessageHandler, condition)

    @classmethod
    def command_handler(cls, condition=None, name=None, prefix="/", min_arguments=None, max_arguments=None):
        return cls.create_event_handler(CommandHandler, condition, name, prefix, min_arguments, max_arguments)

    def on_command(self, condition=None, name=None, prefix="/", min_arguments=None, max_arguments=None):
        return self.add_event_handler(CommandHandler, condition, name, prefix, min_arguments, max_arguments)

    @classmethod
    def callback_query_handler(cls, condition=None):
        return cls.create_event_handler(CallbackQueryHandler, condition)

    def on_callback_query(self, condition=None):
        return self.add_event_handler(CallbackQueryHandler, condition)

    @classmethod
    def pre_checkout_query_handler(cls, condition=None):
        return cls.create_event_handler(PreCheckoutQueryHandler, condition)

    def on_pre_checkout_query(self, condition=None):
        return self.add_event_handler(PreCheckoutQueryHandler, condition)

    @classmethod
    def shipping_query_handler(cls, condition=None):
        return cls.create_event_handler(ShippingQueryHandler, condition)

    def on_shipping_query(self, condition=None):
        return self.add_event_handler(ShippingQueryHandler, condition)

    @classmethod
    def shutdown_handler(cls):
        return cls.create_event_handler(ShutdownHandler)

    def on_shutdown(self):
        return self.add_event_handler(ShutdownHandler)

    @classmethod
    def disconnect_handler(cls):
        return cls.create_event_handler(DisconnectHandler)

    def on_disconnect(self):
        return self.add_event_handler(DisconnectHandler)

    def remove_event_handler(self, event_handler):
        self.children.remove(event_handler)

    def add(self, *chains):
        self.children.extend(chains)

    def include(self, *chains):
        self.chains.extend(chains)

    def get(self, name):
        for chain in self.chains:
            if chain.name == name:
                return chain
        raise ValueError(f"Chain \"{name}\" does not exist")

    def delete(self, name):
        for i, c in enumerate(self.chains):
            if c.name == name:
                del self.chains[i]
                return
        for i, c in enumerate(self.children):
            if isinstance(c, Chain) and c.name == name:
                del self.children[i]
                return
        raise ValueError(f"Chain \"{name}\" does not exist")



import balethon
from ...objects import File


class GetFile:

    async def get_file(
            self: "balethon.Client",
            file_id: str
    ) -> File:
        return await self.auto_execute("get", "getFile", locals())



from typing import Union

import balethon


class CreateNewStickerSet:

    async def create_new_sticker_set(
            self: "balethon.Client",
            user_id: Union[int, str],
            name: str,
            title: str,
            sticker=None
    ) -> bool:
        user_id = await self.resolve_peer_id(user_id)
        return await self.auto_execute("post", "createNewStickerSet", locals())



import balethon


class SetWebhook:

    async def set_webhook(
            self: "balethon.Client",
            url: str
    ) -> bool:
        return await self.auto_execute("post", "setWebhook", locals())



from .create_chat_invite_link import CreateChatInviteLink
from .export_chat_invite_link import ExportChatInviteLink
from .revoke_chat_invite_link import RevokeChatInviteLink


class InviteLinks(CreateChatInviteLink, ExportChatInviteLink, RevokeChatInviteLink):
    pass



from .ban_chat_member import BanChatMember
from .delete_chat_photo import DeleteChatPhoto
from .get_chat import GetChat
from .get_chat_administrators import GetChatAdministrators
from .get_chat_members_count import GetChatMembersCount
from .get_chat_member import GetChatMember
from .leave_chat import LeaveChat
from .pin_chat_message import PinChatMessage
from .promote_chat_member import PromoteChatMember
from .restrict_chat_member import RestrictChatMember
from .send_chat_action import SendChatAction
from .set_chat_description import SetChatDescription
from .set_chat_photo import SetChatPhoto
from .set_chat_title import SetChatTitle
from .unban_chat_member import UnbanChatMember
from .unpin_all_chat_messages import UnpinAllChatMessages
from .unpin_chat_message import UnpinChatMessage


class Chats(
    BanChatMember,
    DeleteChatPhoto,
    GetChat,
    GetChatAdministrators,
    GetChatMembersCount,
    GetChatMember,
    LeaveChat,
    PinChatMessage,
    PromoteChatMember,
    RestrictChatMember,
    SendChatAction,
    SetChatDescription,
    SetChatTitle,
    SetChatPhoto,
    UnbanChatMember,
    UnpinAllChatMessages,
    UnpinChatMessage
):
    pass



from typing import Union

from . import Object
from balethon import objects


class InlineKeyboardButton(Object):

    def __init__(
            self,
            text: str = None,
            callback_data: str = None,
            url: str = None,
            web_app: Union["objects.WebAppInfo", str] = None,
            copy_text: Union["objects.CopyTextButton", str] = None,
            switch_inline_query: str = None,
            switch_inline_query_current_chat: str = None,
            pay: bool = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.text: str = text
        self.callback_data: str = callback_data
        self.url: str = url
        self.web_app = objects.WebAppInfo(web_app) if isinstance(web_app, str) else web_app
        self.copy_text = objects.CopyTextButton(copy_text) if isinstance(copy_text, str) else copy_text
        self.switch_inline_query: str = switch_inline_query
        self.switch_inline_query_current_chat: str = switch_inline_query_current_chat
        self.pay: bool = pay



from typing import Union

import balethon
from ...objects import InviteLink


class CreateChatInviteLink:

    async def create_chat_invite_link(
            self: "balethon.Client",
            chat_id: Union[int, str]
    ) -> InviteLink:
        chat_id = await self.resolve_peer_id(chat_id)
        return await self.auto_execute("post", "createChatInviteLink", locals())



from .condition import create
from ..objects import Message, CallbackQuery
from ..enums import ChatType


@create(can_process=(Message, CallbackQuery))
def private(event) -> bool:
    if isinstance(event, Message):
        return event.chat.type == ChatType.PRIVATE
    elif isinstance(event, CallbackQuery):
        return event.message.chat.type == ChatType.PRIVATE



from datetime import datetime

from .event_handler import EventHandler


class InitializeHandler(EventHandler):

    @property
    def can_handle(self):
        return InitializeHandler

    def __init__(self, callback):
        super().__init__(callback)

    async def handle(self, *args, client=None, event=None, **kwargs):
        if client is not None:
            kwargs["client"] = client
        await super().handle(*args, **kwargs, time=datetime.now())



from .condition import create
from ..objects import Message


@create(can_process=Message)
def media(event) -> bool:
    return bool(event.animation or event.audio or event.document or event.photo or event.video or event.voice)



from . import Object
from balethon import objects
from ..sync_support import add_sync_support_to_object


@add_sync_support_to_object
class CallbackQuery(Object):
    attribute_names = [
        ("author", "from")
    ]

    def __init__(
            self,
            id: str = None,
            author: "objects.User" = None,
            message: "objects.Message" = None,
            inline_message_id: str = None,
            chat_instance: str = None,
            data: str = None,
            game_short_name: str = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.id: str = id
        self.author: "objects.User" = author
        self.message: "objects.Message" = message
        self.inline_message_id: str = inline_message_id
        self.chat_instance: str = chat_instance
        self.data: str = data
        self.game_short_name: str = game_short_name

    async def answer(
            self,
            text: str,
            show_alert: bool = False
    ) -> bool:
        return await self.client.answer_callback_query(self.id, text, show_alert)



from typing import Union, BinaryIO

import balethon
from ...objects import InputMedia, resolve_media, Message, ReplyMarkup


class SendVideo:

    async def send_video(
            self: "balethon.Client",
            chat_id: Union[int, str],
            video: Union[str, bytes, BinaryIO, InputMedia],
            duration: int = None,
            width: int = None,
            height: int = None,
            caption: str = None,
            reply_markup: ReplyMarkup = None,
            reply_to_message_id: int = None
    ) -> Message:
        chat_id = await self.resolve_peer_id(chat_id)
        video = resolve_media(video)
        return await self.auto_execute("post", "sendVideo", locals())



from typing import Union, List

import balethon
from ...objects import ChatMember


class GetChatAdministrators:

    async def get_chat_administrators(
            self: "balethon.Client",
            chat_id: Union[int, str]
    ) -> List[ChatMember]:
        chat_id = await self.resolve_peer_id(chat_id)
        return await self.auto_execute("get", "getChatAdministrators", locals())



from typing import Union, BinaryIO

from . import Object
from balethon import enums, objects
from ..sync_support import add_sync_support_to_object


@add_sync_support_to_object
class Chat(Object):

    def __init__(
            self,
            id: int = None,
            type: "enums.ChatType" = None,
            title: str = None,
            username: str = None,
            first_name: str = None,
            last_name: str = None,
            all_members_are_administrators: bool = None,
            description: str = None,
            invite_link: str = None,
            pinned_message: "objects.Message" = None,
            sticker_set_name: str = None,
            can_set_sticker_set: bool = None,
            photo: "objects.ChatPhoto" = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.id: int = id
        self.type: "enums.ChatType" = type
        self.title: str = title
        self.username: str = username
        self.first_name: str = first_name
        self.last_name: str = last_name
        self.all_members_are_administrators: bool = all_members_are_administrators
        self.description: str = description
        self.invite_link: str = invite_link
        self.pinned_message: "objects.Message" = pinned_message
        self.sticker_set_name: str = sticker_set_name
        self.can_set_sticker_set: bool = can_set_sticker_set
        self.photo: "objects.ChatPhoto" = photo

    def __eq__(self, other):
        return self.id == other.id

    def mention(self, text=None):
        text = text or self.full_name
        return f"[{text}](uid:{self.id})"

    def mention_V2(self, text=None):
        text = text or self.full_name
        return f"[{text}](https://web.bale.ai/chat?uid={self.id})"

    @property
    def link(self):
        if self.invite_link:
            return f"https://{self.invite_link}"
        if self.username:
            return f"{self.client.connection.short_url}/{self.username}"

    @property
    def full_name(self):
        if self.first_name and self.last_name:
            return f"{self.first_name} {self.last_name}"
        if self.first_name:
            return self.first_name
        if self.title:
            return self.title
        return ""

    async def ban_member(self, user_id: Union[int, str]):
        return await self.client.ban_chat_member(self.id, user_id)

    async def delete_photo(self):
        return await self.client.delete_chat_photo(self.id)

    async def get(self):
        return await self.client.get_chat(self.id)

    async def get_administrators(self):
        return await self.client.get_chat_administrators(self.id)

    async def get_member(self, user_id: Union[int, str]):
        return await self.client.get_chat_member(self.id, user_id)

    async def get_members_count(self):
        return await self.client.get_chat_members_count(self.id)

    async def leave(self):
        return await self.client.leave_chat(self.id)

    async def pin_message(self, message_id: int):
        return await self.client.pin_chat_message(self.id, message_id)

    async def promote_member(self, user_id: Union[int, str], *args, **kwargs):
        return await self.client.promote_chat_member(self.id, user_id, *args, **kwargs)

    async def send_action(self, action: "enums.ChatAction" = enums.ChatAction.TYPING):
        return await self.client.send_chat_action(self.id, action)

    async def set_description(self, description: str):
        return await self.client.set_chat_description(self.id, description)

    async def set_photo(self, photo: Union[str, bytes, BinaryIO, "objects.InputMedia"]):
        return await self.client.set_chat_photo(self.id, photo)

    async def set_title(self, title: str):
        return await self.client.set_chat_title(self.id, title)

    async def unban_member(self, user_id: Union[int, str], only_if_banned: bool = True):
        return await self.client.unban_chat_member(self.id, user_id, only_if_banned)

    async def unpin_message(self, message_id: int = None):
        return await self.client.unpin_chat_message(self.id, message_id)



from . import Object


class ReplyMarkup(Object):

    @classmethod
    def wrap(cls, raw_object):
        from . import InlineKeyboard, ReplyKeyboard
        if "inline_keyboard" in raw_object:
            return InlineKeyboard.wrap(raw_object)
        elif "keyboard" in raw_object:
            return ReplyKeyboard.wrap(raw_object)



import balethon


class AnswerCallbackQuery:

    async def answer_callback_query(
            self: "balethon.Client",
            callback_query_id: str,
            text: str,
            show_alert: bool = False
    ) -> bool:
        return await self.auto_execute("post", "answerCallbackQuery", locals())



import balethon


class AnswerPreCheckoutQuery:

    async def answer_pre_checkout_query(
            self: "balethon.Client",
            pre_checkout_query_id: str,
            ok: bool = True,
            error_message: str = None
    ) -> bool:
        return await self.auto_execute("post", "answerPreCheckoutQuery", locals())



from . import Object
from balethon import enums, objects


class ChatMember(Object):

    def __init__(
            self,
            is_member: bool = None,
            can_send_media: bool = None,
            can_send_gif_stickers: bool = None,
            can_forward_message_from: bool = None,
            can_send_gift_packet: bool = None,
            can_send_link_message: bool = None,
            can_send_forwarded_message: bool = None,
            can_kick_user: bool = None,
            can_send_message: bool = None,
            can_see_members: bool = None,
            can_add_story: bool = None,
            user: "objects.User" = None,
            status: "enums.ChatMemberStatus" = None,
            can_change_info: bool = None,
            can_delete_messages: bool = None,
            can_invite_users: bool = None,
            can_restrict_members: bool = None,
            can_pin_messages: bool = None,
            can_promote_members: bool = None,
            can_send_messages: bool = None,
            can_send_media_messages: bool = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.is_member: bool = is_member
        self.can_send_media: bool = can_send_media
        self.can_send_gif_stickers: bool = can_send_gif_stickers
        self.can_forward_message_from: bool = can_forward_message_from
        self.can_send_gift_packet: bool = can_send_gift_packet
        self.can_send_link_message: bool = can_send_link_message
        self.can_send_forwarded_message: bool = can_send_forwarded_message
        self.can_kick_user: bool = can_kick_user
        self.can_send_message: bool = can_send_message
        self.can_see_members: bool = can_see_members
        self.can_add_story: bool = can_add_story
        self.user: "objects.User" = user
        self.status: "enums.ChatMemberStatus" = status
        self.can_change_info: bool = can_change_info
        self.can_delete_messages: bool = can_delete_messages
        self.can_invite_users: bool = can_invite_users
        self.can_restrict_members: bool = can_restrict_members
        self.can_pin_messages: bool = can_pin_messages
        self.can_promote_members: bool = can_promote_members
        self.can_send_messages: bool = can_send_messages
        self.can_send_media_messages: bool = can_send_media_messages



from logging import getLogger

from .chain import Chain


class LoggingChain(Chain):

    def __init__(self, name="logging", condition=None, logger=None):
        super().__init__(name, condition)
        self.log = logger or getLogger(self.name)

    @Chain.error_handler()
    def log_error(self, error):
        self.log.exception(error)

    @Chain.connect_handler()
    def log_connect(self, client):
        self.log.info(f"{client} connected")

    @Chain.initialize_handler()
    def log_initialize(self, client):
        sync_workers = client.dispatcher.sync_workers_count
        async_workers = client.dispatcher.async_workers_count
        self.log.info(f"{client} initialized (sync workers: {sync_workers} - async workers: {async_workers})")

    @Chain.shutdown_handler()
    def log_shutdown(self, client):
        self.log.info(f"{client} shutting down")

    @Chain.disconnect_handler()
    def log_disconnect(self, client):
        self.log.info(f"{client} disconnected")

    @Chain.event_handler()
    def log_event(self, event):
        self.log.info(event)



from .condition import Condition


class IsInstance(Condition):
    def __init__(self, classes):
        super().__init__()
        self.classes = classes

    async def __call__(self, client, event) -> bool:
        return isinstance(event, self.classes)



from typing import get_type_hints, get_args, get_origin, Union, List, Optional
from copy import copy
from json import dumps

from ..enums import NameEnum
import balethon


class Object:
    attribute_names = []

    @classmethod
    def expected_types(cls):
        return get_type_hints(cls.__init__)

    @classmethod
    def validate_types(cls, raw_object):
        expected_types = cls.expected_types()
        for key, value in raw_object.items():
            if not expected_types.get(key):
                continue
            if value is None:
                continue
            expected_type = expected_types[key]
            if isinstance(expected_type, (type(Union), type(Union[str, int]))) and get_origin(expected_type) in (Union, None):
                expected_type = get_args(expected_type)[0]
            if isinstance(expected_type, (type(List), type(List[str]))) and get_origin(expected_type) == list:
                if isinstance(value, list):
                    raw_object[key] = wrap(expected_type, value)
                continue
            if issubclass(expected_type, NameEnum):
                try:
                    raw_object[key] = expected_type(value)
                except ValueError:
                    pass
                else:
                    continue
            if not issubclass(expected_type, Object):
                continue
            if isinstance(value, expected_type):
                continue
            raw_object[key] = expected_type.wrap(value)
        return raw_object

    @classmethod
    def wrap(cls, raw_object):
        for attribute_name, key_name in cls.attribute_names:
            if raw_object.get(key_name):
                raw_object[attribute_name] = raw_object.pop(key_name)
        raw_object = cls.validate_types(raw_object)
        return cls(**raw_object)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __getitem__(self, item):
        return getattr(self, item)

    def bind(self, client):
        self.client = client
        for value in self.__dict__.values():
            if isinstance(value, Object):
                value.bind(client)

    def __init__(
            self,
            **kwargs
    ):
        self.client: Optional["balethon.Client"] = None
        for key, value in kwargs.items():
            self[key] = value

    def unwrap(self):
        result = copy(self)
        del result.client
        for key, value in result.__dict__.copy().items():
            if isinstance(value, NameEnum):
                result[key] = value.value
            if isinstance(value, list):
                result[key] = unwrap(value)
            if isinstance(value, Object):
                result[key] = value.unwrap()
            if value is None:
                delattr(result, key)
        for attribute_name, key_name in result.attribute_names:
            if result.__dict__.get(attribute_name):
                result[key_name] = result.__dict__.pop(attribute_name)
        return result.__dict__

    def to_json(self):
        return dumps(self.unwrap(), ensure_ascii=False, indent=4)

    def __repr__(self):
        attributes = []
        for key, value in self.__dict__.items():
            if key == "client":
                continue
            if value is None:
                continue
            attributes.append(f"{key}={repr(value)}")

        if not attributes:
            return f"{type(self).__name__}()"

        attributes = ",\n".join(attributes)
        attributes = "\n".join(" "*4 + line for line in attributes.splitlines())
        return f"{type(self).__name__}(\n{attributes}\n)"


def wrap(expected_type, raw_object):
    if isinstance(expected_type, (type(List), type(List[str]))) and get_origin(expected_type) == list:
        if isinstance(raw_object, list):
            expected_type = get_args(expected_type)[0]
            raw_object = copy(raw_object)
            for i, element in enumerate(raw_object):
                if element is None:
                    continue
                if isinstance(expected_type, (type(List), type(List[str]))) and get_origin(expected_type) == list:
                    if isinstance(element, list):
                        raw_object[i] = wrap(expected_type, element)
                    continue
                if not issubclass(expected_type, Object):
                    continue
                if isinstance(element, expected_type):
                    continue
                raw_object[i] = expected_type.wrap(element)
            from .list import List as BalethonList
            return BalethonList(raw_object)

    if issubclass(expected_type, Object):
        return expected_type.wrap(raw_object)

    return raw_object


def unwrap(wrapped_object):
    if isinstance(wrapped_object, list):
        wrapped_object = copy(wrapped_object)
        for i, element in enumerate(wrapped_object):
            wrapped_object[i] = unwrap(element)
        return list(wrapped_object)

    if isinstance(wrapped_object, Object):
        return wrapped_object.unwrap()
    
    if isinstance(wrapped_object, NameEnum):
        return wrapped_object.value

    return wrapped_object



from . import ReplyKeyboard


class ReplyKeyboardRemove(ReplyKeyboard):

    def __init__(self):
        super().__init__(remove=True)

    def __repr__(self):
        return f"{type(self).__name__}()"



from typing import Union, BinaryIO

import balethon
from ...objects import InputMedia, resolve_media, Message, ReplyMarkup


class SendAnimation:

    async def send_animation(
            self: "balethon.Client",
            chat_id: Union[int, str],
            animation: Union[str, bytes, BinaryIO, InputMedia],
            duration: int = None,
            width: int = None,
            height: int = None,
            caption: str = None,
            reply_markup: ReplyMarkup = None,
            reply_to_message_id: int = None
    ) -> Message:
        chat_id = await self.resolve_peer_id(chat_id)
        animation = resolve_media(animation)
        return await self.auto_execute(
            "post",
            "sendAnimation",
            locals(),
            json=False,  # The sendAnimation method only works with multipart/form-data
        )



from . import Object, unwrap


class List(list, Object):
    @classmethod
    def wrap(cls, raw_object):
        return cls(raw_object)

    def bind(self, client):
        self.client = client
        for element in self:
            if isinstance(element, Object):
                element.bind(client)

    def unwrap(self):
        return unwrap(self)

    def __repr__(self):
        if not self:
            return "[]"

        elements = ",\n".join(f"{repr(element)}" for element in self)
        elements = "\n".join(" "*4 + line for line in elements.splitlines())
        return f"[\n{elements}\n]"



from . import Object


class WebAppData(Object):
    
    def __init__(
            self,
            data: str = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.data: str = data



from . import Object
from balethon import objects


class OrderInfo(Object):

    def __init__(
            self,
            name: str = None,
            phone_number: str = None,
            email: str = None,
            shipping_address: "objects.ShippingAddress" = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.name: str = name
        self.phone_number: str = phone_number
        self.email: str = email
        self.shipping_address: "objects.ShippingAddress" = shipping_address



from json import loads
from re import search
from logging import getLogger

from httpx import AsyncClient
from httpx._types import ProxyTypes

from ..errors import RPCError

log = getLogger(__name__)


class Connection:
    TIMEOUT = 20
    BASE_URL = "https://tapi.bale.ai"
    SHORT_URL = "https://ble.ir"

    def __init__(
            self,
            token: str,
            time_out: int = None,
            proxy: ProxyTypes = None,
            base_url: str = None,
            short_url: str = None
    ):
        self.token = token
        self.client = None
        self.is_started = False
        self.time_out = time_out or self.TIMEOUT
        self.proxy = proxy
        self.base_url = base_url or self.BASE_URL
        self.short_url = short_url or self.SHORT_URL

    async def start(self):
        if self.is_started:
            raise ConnectionError("Connection is already started")
        self.is_started = True
        self.client = AsyncClient(proxy=self.proxy)

    async def stop(self):
        if not self.is_started:
            raise ConnectionError("Connection is already stopped")
        self.is_started = False
        await self.client.aclose()
        self.client = None

    def bot_url(self) -> str:
        return f"{self.base_url}/bot{self.token}"

    def file_url(self, file_id: str) -> str:
        return f"{self.base_url}/file/bot{self.token}/{file_id}"

    async def get_peer_info(self, query: str):
        chat_id = query if query.startswith("@") else f"@{query}"
        info = await self.client.post(f"{self.bot_url()}/getChat?chat_id={chat_id}")
        info = info.json()
        if info.get("ok", False):
            return ["v1", info]
        response = await self.client.get(f"{self.short_url}/{query}")
        json_info = search(r'(<script id="__NEXT_DATA__" type="application/json">.*</script>)', response.text)[0]
        json_info = search(r"({.*})", json_info)[0]
        return ["v2", loads(json_info)]

    async def request(self, method: str, service: str, data: dict = None, json: dict = None, files: dict = None):
        if json:
            log.info(f"[{service}] JSON{json}")
        if data:
            log.info(f"[{service}] DATA{data} - FILES{files}")
        response = await self.client.request(
            method,
            f"{self.bot_url()}/{service}",
            data=data,
            files=files,
            json=json,
            timeout=self.time_out
        )
        response_json = response.json()
        if response.status_code != 200:
            code = response.status_code or response_json.get("error_code")
            raise RPCError.create(code, response_json.get("description"), service, response_json.get("parameters"))
        return response_json.get("result")

    async def download_file(self, file_id: str):
        response = await self.client.get(self.file_url(file_id))
        if response.status_code != 200:
            response_json = response.json()
            raise RPCError.create(response.status_code, response_json.get("description"), "downloadFile")
        return response.read()



from .condition import Condition
from ..objects import Message, CallbackQuery


class Equals(Condition):
    def __init__(self, *values):
        super().__init__(can_process=(Message, CallbackQuery))
        self.values = values

    async def __call__(self, client, event) -> bool:
        if isinstance(event, Message):
            return event.content in self.values
        elif isinstance(event, CallbackQuery):
            return event.data in self.values



from . import Object
from balethon import objects


class Entity(Object):

    def __init__(
            self,
            type: str = None,
            offset: int = None,
            length: int = None,
            url: str = None,
            user: "objects.User" = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.type: str = type
        self.offset: int = offset
        self.length: int = length
        self.url: str = url
        self.user: "objects.User" = user



from .send_photo import SendPhoto
from .send_animation import SendAnimation
from .send_audio import SendAudio
from .send_document import SendDocument
from .send_sticker import SendSticker
from .send_video import SendVideo
from .send_voice import SendVoice
from .send_location import SendLocation
from .send_contact import SendContact
from .send_media_group import SendMediaGroup
from .get_file import GetFile


class Attachments(
    SendPhoto,
    SendAnimation,
    SendAudio,
    SendDocument,
    SendVideo,
    SendVoice,
    SendSticker,
    SendLocation,
    SendContact,
    SendMediaGroup,
    GetFile
):
    pass



from typing import Union

import balethon


class ExportChatInviteLink:

    async def export_chat_invite_link(
            self: "balethon.Client",
            chat_id: Union[int, str]
    ) -> str:
        chat_id = await self.resolve_peer_id(chat_id)
        return await self.auto_execute("post", "exportChatInviteLink", locals())



from . import Object


class ChatPhoto(Object):
    attribute_names = [
        ("small_id", "small_file_id"),
        ("small_unique_id", "small_file_unique_id"),
        ("big_id", "big_file_id"),
        ("big_unique_id", "big_file_unique_id")
    ]

    def __init__(
            self,
            small_id: str = None,
            small_unique_id: str = None,
            big_id: str = None,
            big_unique_id: str = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.small_id: str = small_id
        self.small_unique_id: str = small_unique_id
        self.big_id: str = big_id
        self.big_unique_id: str = big_unique_id



from typing import Union

import balethon


class UnbanChatMember:

    async def unban_chat_member(
            self: "balethon.Client",
            chat_id: Union[int, str],
            user_id: Union[int, str],
            only_if_banned: bool = True
    ) -> bool:
        chat_id = await self.resolve_peer_id(chat_id)
        user_id = await self.resolve_peer_id(user_id)
        return await self.auto_execute("post", "unbanChatMember", locals())



from inspect import iscoroutinefunction

from ..smart_call import remove_unwanted_keyword_parameters


class Condition:
    def __init__(self, function=None, can_process=None):
        self.function = function
        self.can_process = can_process

    async def __call__(self, client, event) -> bool:
        if self.can_process is not None and not isinstance(event, self.can_process):
            return False
        kwargs = dict(condition=self, client=client, event=event)
        kwargs = remove_unwanted_keyword_parameters(self.function, **kwargs)
        if iscoroutinefunction(self.function):
            return await self.function(**kwargs)
        return self.function(**kwargs)

    def __and__(self, other):
        return AllCondition(self, other)

    def __or__(self, other):
        return AnyCondition(self, other)

    def __invert__(self):
        return NotCondition(self)

    def __repr__(self):
        return type(self).__name__


class AllCondition(Condition):
    def __init__(self, *conditions):
        super().__init__()
        self.conditions = conditions

    async def __call__(self, client, event) -> bool:
        for condition in self.conditions:
            if not await condition(client, event):
                return False
        return True

    def __repr__(self):
        conditions_string = ", ".join(map(str, self.conditions))
        return f"All({conditions_string})"


class AnyCondition(Condition):
    def __init__(self, *conditions):
        super().__init__()
        self.conditions = conditions

    async def __call__(self, client, event) -> bool:
        for condition in self.conditions:
            if await condition(client, event):
                return True
        return False

    def __repr__(self):
        conditions_string = ", ".join(map(str, self.conditions))
        return f"Any({conditions_string})"


class NotCondition(Condition):
    def __init__(self, condition):
        super().__init__()
        self.condition = condition

    async def __call__(self, client, event) -> bool:
        return not await self.condition(client, event)

    def __repr__(self):
        return f"Not({self.condition})"


def create(name: str = None, can_process=None):
    def decorator(function):
        CustomCondition = type(name or function.__name__ or "CustomCondition", (Condition,), {})
        return CustomCondition(function, can_process)

    return decorator



from enum import auto

from .name_enum import NameEnum


class MessageMediaType(NameEnum):
    PHOTO = auto()
    VOICE = auto()
    AUDIO = auto()
    VIDEO = auto()
    ANIMATION = auto()
    CONTACT = auto()
    LOCATION = auto()
    STICKER = auto()
    DOCUMENT = auto()



from .update_handler import UpdateHandler
from ..objects import ShippingQuery


class ShippingQueryHandler(UpdateHandler):
    can_handle = ShippingQuery

    def __init__(self, callback, condition=None):
        super().__init__(callback, condition)

    def handle(self, *args, client=None, event=None, **kwargs):
        if client is not None:
            kwargs["client"] = client
        if event is not None:
            kwargs["shipping_query"] = event
        return super().handle(*args, **kwargs)



from typing import Union, BinaryIO

import balethon
from ...objects import InputMedia, resolve_media


class SetChatPhoto:

    async def set_chat_photo(
            self: "balethon.Client",
            chat_id: Union[int, str],
            photo: Union[str, bytes, BinaryIO, InputMedia]
    ) -> bool:
        chat_id = await self.resolve_peer_id(chat_id)
        photo = resolve_media(photo)
        return await self.auto_execute("post", "setChatPhoto", locals())



import balethon
from ...objects import StickerSet


class GetStickerSet:

    async def get_sticker_set(
            self: "balethon.Client",
            name: str
    ) -> StickerSet:
        return await self.auto_execute("get", "getStickerSet", locals())



import balethon
from ...objects import WebhookInfo


class GetWebhookInfo:

    async def get_webhook_info(
            self: "balethon.Client"
    ) -> WebhookInfo:
        return await self.auto_execute("post", "getWebhookInfo", locals())



from typing import Union, BinaryIO

import balethon
from ...objects import InputMedia, resolve_media


class UploadStickerFile:

    async def upload_sticker_file(
            self: "balethon.Client",
            user_id: Union[int, str],
            sticker: Union[str, bytes, BinaryIO, InputMedia]
    ) -> str:
        user_id = await self.resolve_peer_id(user_id)
        sticker = resolve_media(sticker)
        return await self.auto_execute("get", "uploadStickerFile", locals())



from typing import Union

import balethon
from ...objects import Message, ReplyMarkup


class SendLocation:

    async def send_location(
            self: "balethon.Client",
            chat_id: Union[int, str],
            latitude: int,
            longitude: int,
            reply_markup: ReplyMarkup = None,
            reply_to_message_id: int = None
    ) -> Message:
        chat_id = await self.resolve_peer_id(chat_id)
        return await self.auto_execute("post", "sendLocation", locals())



from . import Object
from balethon import objects


class Update(Object):
    attribute_names = [
        ("id", "update_id")
    ]

    def __init__(
            self,
            id: int = None,
            message: "objects.Message" = None,
            edited_message: "objects.EditedMessage" = None,
            channel_post: "objects.Message" = None,
            edited_channel_post: "objects.EditedMessage" = None,
            callback_query: "objects.CallbackQuery" = None,
            shipping_query: "objects.ShippingQuery" = None,
            pre_checkout_query: "objects.PreCheckoutQuery" = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.id: int = id
        self.message: "objects.Message" = message
        self.edited_message: "objects.EditedMessage" = edited_message
        self.channel_post: "objects.Message" = channel_post
        self.edited_channel_post: "objects.EditedMessage" = edited_channel_post
        self.callback_query: "objects.CallbackQuery" = callback_query
        self.shipping_query: "objects.ShippingQuery" = shipping_query
        self.pre_checkout_query: "objects.PreCheckoutQuery" = pre_checkout_query

    def get_effective_update(self):
        for value in self.__dict__.values():
            if isinstance(value, Object):
                return value



from typing import Callable
from inspect import iscoroutinefunction
from functools import partial

from ..conditions import Condition, create
from ..smart_call import remove_unwanted_parameters


class EventHandler:
    can_handle = object

    def __init__(self, callback: Callable, condition=None):
        self.callback = callback
        if not isinstance(condition, Condition) and condition is not None:
            condition = create(condition)
        self.condition = condition
        self.self = None

    def to_string(self, keyword="if", tabs=0):
        result = []
        condition = self.condition or self.can_handle.__name__
        result.append(f"{keyword} {condition}:")
        result.append(f"    {type(self).__name__}({self.can_handle.__name__})")
        return "\n".join(f"{'    ' * tabs}{i}" for i in result)

    def __repr__(self):
        return self.to_string()

    async def check(self, client, event):
        if self.condition is None:
            return True
        return await self.condition(client, event)

    async def handle(self, *args, client=None, event=None, **kwargs):
        if client is not None:
            kwargs["client"] = client
        client = kwargs["client"]
        if self.self is not None:
            args = list(args)
            args.insert(0, self.self)
        if event is not None:
            kwargs["event"] = event
        args, kwargs = remove_unwanted_parameters(self.callback, *args, **kwargs)
        if iscoroutinefunction(self.callback):
            return await self.callback(*args, **kwargs)
        return client.dispatcher.event_loop.run_in_executor(
            client.dispatcher.sync_workers,
            partial(self.callback, *args, **kwargs)
        )

    def __call__(self, *args, **kwargs):
        args, kwargs = remove_unwanted_parameters(self.callback, *args, **kwargs)
        return self.callback(*args, **kwargs)



from .delete_webhook import DeleteWebhook
from .get_updates import GetUpdates
from .get_webhook_info import GetWebhookInfo
from .set_webhook import SetWebhook


class Updates(DeleteWebhook, GetUpdates, GetWebhookInfo, SetWebhook):
    pass



from .rpc_error import RPCError


class TooManyRequestsError(RPCError):
    name = "Too Many Requests"
    code = 429

    def __init__(self, code: int = None, description: str = None, reason: str = None, parameters: dict = None):
        super().__init__(code, description, reason, parameters)
        self.seconds = parameters.get("retry_after", 0)



from typing import Union

import balethon


class LeaveChat:

    async def leave_chat(
            self: "balethon.Client",
            chat_id: Union[int, str]
    ) -> bool:
        chat_id = await self.resolve_peer_id(chat_id)
        return await self.auto_execute("post", "leaveChat", locals())



from . import Object


class WebhookInfo(Object):

    def __init__(
            self,
            url: str,
            has_custom_certificate: bool,
            pending_update_count: int,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.url: str = url
        self.has_custom_certificate: bool = has_custom_certificate
        self.pending_update_count: int = pending_update_count



from .dispatching_errors import BreakDispatching, ContinueDispatching
from .rpc_errors import (
    BadRequestError,
    FloodError,
    ForbiddenError,
    InternalError,
    NotFoundError,
    RPCError,
    UnauthorizedError,
    TooManyRequestsError
)



from . import Object
from balethon import objects


class SuccessfulPayment(Object):

    def __init__(
            self,
            currency: str = None,
            total_amount: int = None,
            invoice_payload: str = None,
            shipping_option_id: str = None,
            order_info: "objects.OrderInfo" = None,
            telegram_payment_charge_id: str = None,
            provider_payment_charge_id: str = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.currency: str = currency
        self.total_amount: int = total_amount
        self.invoice_payload: str = invoice_payload
        self.shipping_option_id: str = shipping_option_id
        self.order_info: "objects.OrderInfo" = order_info
        self.telegram_payment_charge_id: str = telegram_payment_charge_id
        self.provider_payment_charge_id: str = provider_payment_charge_id

    async def inquire(
            self
    ):
        return await self.client.inquire_transaction(self.telegram_payment_charge_id)



from inspect import getfullargspec

from .message_handler import MessageHandler
from ..conditions import command
from ..objects import Message


class CommandHandler(MessageHandler):
    can_handle = Message

    @staticmethod
    def get_min_arguments(callback):
        args, _, __, defaults, *___ = getfullargspec(callback)
        args_count = len(args)
        defaults_count = 0 if defaults is None else len(defaults)
        return args_count - defaults_count

    @staticmethod
    def get_max_arguments(callback):
        args, varargs, *_ = getfullargspec(callback)
        if varargs is not None:
            return None
        return len(args)

    def __init__(self, callback, condition=None, name=None, prefix="/", min_arguments=None, max_arguments=None):
        if name is None:
            name = callback.__name__
        if min_arguments is None:
            min_arguments = self.get_min_arguments(callback)
        if max_arguments is None:
            max_arguments = self.get_max_arguments(callback)
        command_condition = command(name, prefix, min_arguments, max_arguments)
        if condition is None:
            condition = command_condition
        else:
            condition = command_condition & condition
        super().__init__(callback, condition)

    def handle(self, *args, client=None, event=None, **kwargs):
        if client is not None:
            kwargs["client"] = client
        if event is not None:
            kwargs["message"] = event
            _, *arguments = event.text.split()
            args += tuple(arguments)
        return super().handle(*args, **kwargs)



from .object import Object, wrap, unwrap
from .user import User
from .date import Date
from .chat import Chat
from .update import Update
from .file import File
from .chat_member import ChatMember
from .message import Message
from .edited_message import EditedMessage
from .callback_query import CallbackQuery
from .reply_markup import ReplyMarkup
from .inline_keyboard_button import InlineKeyboardButton
from .inline_keyboard import InlineKeyboard
from .reply_keyboard_button import ReplyKeyboardButton
from .reply_keyboard import ReplyKeyboard
from .entity import Entity
from .shipping_query import ShippingQuery
from .shipping_address import ShippingAddress
from .pre_checkout_query import PreCheckoutQuery
from .order_info import OrderInfo
from .labeled_price import LabeledPrice
from .audio import Audio
from .contact import Contact
from .document import Document
from .location import Location
from .photo import Photo
from .video import Video
from .voice import Voice
from .invoice import Invoice
from .successful_payment import SuccessfulPayment
from .transaction import Transaction
from .input_media import InputMedia, resolve_media
from .input_media_photo import InputMediaPhoto
from .input_media_video import InputMediaVideo
from .sticker import Sticker
from .sticker_set import StickerSet
from .reply_keyboard_remove import ReplyKeyboardRemove
from .animation import Animation
from .webhook_info import WebhookInfo
from .chat_photo import ChatPhoto
from .invite_link import InviteLink
from .list import List
from .web_app_data import WebAppData
from .web_app_info import WebAppInfo
from .copy_text_button import CopyTextButton



from .condition import Condition, AllCondition, AnyCondition, NotCondition, create
from .all import all
from .forward import forward
from .reply import reply
from .text import text
from .entities import entities
from .document import document
from .photo import photo
from .video import video
from .voice import voice
from .caption import caption
from .contact import contact
from .location import location
from .new_chat_members import new_chat_members
from .left_chat_member import left_chat_member
from .new_chat_title import new_chat_title
from .new_chat_photo import new_chat_photo
from .delete_chat_photo import delete_chat_photo
from .group_chat_created import group_chat_created
from .supergroup_chat_created import supergroup_chat_created
from .channel_chat_created import channel_chat_created
from .pinned_message import pinned_message
from .invoice import invoice
from .media import media
from .command import Command as command
from .private import private
from .group import group
from .channel import channel
from .regex import Regex as regex
from .at_state import AtState as at_state
from .is_instance import IsInstance as is_instance
from .chat import Chat as chat
from .is_joined import IsJoined as is_joined
from .author import Author as author
from .equals import Equals as equals
from .successful_payment import successful_payment
from .content import content
from .media_group import media_group
from .sticker import sticker
from .number import number
from .click import Click as click
from .audio import audio
from .animation import animation



from .condition import Condition
from ..objects import Message, CallbackQuery, ReplyKeyboard, InlineKeyboard


class Click(Condition):
    def __init__(self, keyboard, row_index: int, button_index: int):
        super().__init__(can_process=(Message, CallbackQuery))
        self.keyboard = keyboard
        self.row_index = row_index
        self.button_index = button_index

    async def __call__(self, client, event) -> bool:
        if isinstance(self.keyboard, ReplyKeyboard) and isinstance(event, Message):
            button = self.keyboard.keyboard[self.row_index][self.button_index]
            return event.text == button.text
        elif isinstance(self.keyboard, InlineKeyboard) and isinstance(event, CallbackQuery):
            button = self.keyboard.inline_keyboard[self.row_index][self.button_index]
            return event.data == button.callback_data



from . import Object
from balethon import objects


class InviteLink(Object):
    attribute_names = [
        ("link", "invite_link")
    ]

    def __init__(
            self,
            link: str = None,
            creator: "objects.User" = None,
            creates_join_request: bool = None,
            is_primary: bool = None,
            is_revoked: bool = None,
            name: str = None,
            expire_date: int = None,
            member_limit: int = None,
            pending_join_request_count: int = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.link: str = link
        self.creator: "objects.User" = creator
        self.creates_join_request: bool = creates_join_request
        self.is_primary: bool = is_primary
        self.is_revoked: bool = is_revoked
        self.name: str = name
        self.expire_date: int = expire_date
        self.member_limit: int = member_limit
        self.pending_join_request_count: int = pending_join_request_count



from datetime import datetime

from . import Object


class Date(Object, datetime):

    @classmethod
    def wrap(cls, raw_object):
        date_time = datetime.fromtimestamp(raw_object)
        return super().__new__(cls, *date_time.timetuple()[:6])

    def __new__(
            cls,
            *args,
            **kwargs
    ):
        return super().__new__(cls, *args, *kwargs)

    def __init__(
            self,
            *args,
            **kwargs
    ):
        super().__init__(**kwargs)

    def __repr__(self):
        return str(self)

    def unwrap(self):
        return int(super().timestamp())



from .condition import Condition
from ..objects import Message, CallbackQuery


class Author(Condition):
    def __init__(self, *authors):
        super().__init__(can_process=(Message, CallbackQuery))
        self.authors = set(authors)

    async def __call__(self, client, event) -> bool:
        if not event.author:
            return False
        return event.author.id in self.authors



from functools import wraps
from asyncio import get_event_loop, new_event_loop, set_event_loop, wrap_future, run_coroutine_threadsafe
from inspect import iscoroutinefunction
from threading import current_thread, main_thread
from typing import Callable


def add_sync_support_to_function(coroutine_function) -> Callable:
    main_loop = get_event_loop()

    @wraps(coroutine_function)
    def dual_purpose_function(*args, **kwargs):
        coroutine = coroutine_function(*args, **kwargs)
        try:
            loop = get_event_loop()
        except RuntimeError:
            loop = new_event_loop()
            set_event_loop(loop)
        if current_thread() is main_thread() or not main_loop.is_running():
            if loop.is_running():
                return coroutine
            return loop.run_until_complete(coroutine)
        if loop.is_running():
            async def coroutine_wrapper():
                return await wrap_future(run_coroutine_threadsafe(coroutine, main_loop))
            return coroutine_wrapper()
        return run_coroutine_threadsafe(coroutine, main_loop).result()

    return dual_purpose_function


def add_sync_support_to_object(obj):
    for name in dir(obj):
        attribute = getattr(obj, name)

        if name.startswith("_"):
            continue
        if not iscoroutinefunction(attribute):
            continue

        dual_purpose_method = add_sync_support_to_function(attribute)
        setattr(obj, name, dual_purpose_method)

    return obj



from typing import Union

import balethon


class GetChatMembersCount:

    async def get_chat_members_count(
            self: "balethon.Client",
            chat_id: Union[int, str]
    ) -> int:
        chat_id = await self.resolve_peer_id(chat_id)
        return await self.auto_execute("get", "getChatMembersCount", locals())



from . import Object
from balethon import objects


class PreCheckoutQuery(Object):
    attribute_names = [
        ("author", "from")
    ]

    def __init__(
            self,
            id: str = None,
            author: "objects.User" = None,
            currency: str = None,
            total_amount: int = None,
            invoice_payload: str = None,
            shipping_option_id: str = None,
            order_info: "objects.OrderInfo" = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.id: str = id
        self.author: "objects.User" = author
        self.currency: str = currency
        self.total_amount: int = total_amount
        self.invoice_payload: str = invoice_payload
        self.shipping_option_id: str = shipping_option_id
        self.order_info: "objects.OrderInfo" = order_info

    async def answer(
            self,
            ok: bool = True,
            error_message: str = None
    ) -> bool:
        return await self.client.answer_pre_checkout_query(self.id, ok, error_message)

    async def inquire(
            self
    ):
        return await self.client.inquire_transaction(self.id)



from .answer_callback_query import AnswerCallbackQuery
from .send_message import SendMessage
from .edit_message_caption import EditMessageCaption
from .edit_message_reply_markup import EditMessageReplyMarkup
from .edit_message_text import EditMessageText
from .forward_message import ForwardMessage
from .delete_message import DeleteMessage
from .copy_message import CopyMessage


class Messages(
    AnswerCallbackQuery,
    SendMessage,
    EditMessageCaption,
    EditMessageReplyMarkup,
    EditMessageText,
    ForwardMessage,
    DeleteMessage,
    CopyMessage
):
    pass



from typing import Union

import balethon
from balethon import enums


class SendChatAction:

    async def send_chat_action(
            self: "balethon.Client",
            chat_id: Union[int, str],
            action: "enums.ChatAction" = enums.ChatAction.TYPING
    ) -> bool:
        chat_id = await self.resolve_peer_id(chat_id)
        chat_id = str(chat_id)  # The sendChatAction method only works with a string chat_id
        return await self.auto_execute("post", "sendChatAction", locals())



from typing import Union

import balethon


class PromoteChatMember:

    async def promote_chat_member(
            self: "balethon.Client",
            chat_id: Union[int, str],
            user_id: Union[int, str],
            can_edit_messages: bool = None,
            can_delete_messages: bool = None,
            can_manage_video_chats: bool = None,
            can_restrict_members: bool = None,
            can_promote_members: bool = None,
            can_change_info: bool = None,
            can_invite_users: bool = None,
            can_pin_messages: bool = None,
            can_send_messages: bool = None,
            can_send_media_messages: bool = None,
            can_send_media: bool = None,
            can_send_gif_stickers: bool = None,
            can_reply_to_story: bool = None,
            can_forward_message_from: bool = None,
            can_send_gift_packet: bool = None,
            can_start_call: bool = None,
            can_send_link_message: bool = None,
            can_send_forwarded_message: bool = None,
            can_kick_user: bool = None,
            can_send_message: bool = None,
            can_see_members: bool = None,
            can_add_story: bool = None
    ) -> bool:
        chat_id = await self.resolve_peer_id(chat_id)
        user_id = await self.resolve_peer_id(user_id)
        return await self.auto_execute("post", "promoteChatMember", locals())



from .event_handler import EventHandler
from ..objects import Object


class UpdateHandler(EventHandler):
    can_handle = Object

    def __init__(self, callback, condition=None):
        super().__init__(callback, condition)

    async def handle(self, *args, client=None, event=None, **kwargs):
        if client is not None:
            kwargs["client"] = client
        if event is not None:
            kwargs["update"] = event
        await super().handle(*args, **kwargs)
